from uuid import UUID
from rest_framework import serializers
from django.conf import settings
from django.db.models import Q
from ponos.models import State
from arkindex_common.enums import DataImportMode
from arkindex.project.serializer_fields import EnumField, BestClassField
from arkindex.documents.models import Corpus, Element, ElementType, Classification, ClassificationState
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.dataimport.models import DataImport, DataFile, WorkerRun, WorkerVersion
from arkindex.dataimport.serializers.git import RevisionSerializer
from transkribus import TranskribusAPI


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(State)
    mode = EnumField(DataImportMode, read_only=True)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    workflow = serializers.HyperlinkedRelatedField(read_only=True, view_name='ponos:workflow-details')

    class Meta:
        model = DataImport
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
            'workflow',
        )
        read_only_fields = ('id', 'state', 'mode', 'corpus', 'creator', 'workflow')


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its settings
    """

    # Redefine state as writable
    state = EnumField(State, read_only=True)
    revision = RevisionSerializer(read_only=True)
    element = ElementSlimSerializer(read_only=True)
    folder_type = serializers.SlugField(source='folder_type.slug', default=None, read_only=True)
    element_type = serializers.SlugRelatedField(queryset=ElementType.objects.none(), slug_field='slug', allow_null=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        dataimport = self.context.get('dataimport')
        if not dataimport or not dataimport.corpus:
            return
        self.fields['element_type'].queryset = ElementType.objects.filter(corpus=dataimport.corpus)

    class Meta(DataImportLightSerializer.Meta):
        fields = DataImportLightSerializer.Meta.fields + (
            'files',
            'revision',
            'element',
            'folder_type',
            'element_type',
            'name_contains',
            'load_children'
        )
        read_only_fields = DataImportLightSerializer.Meta.read_only_fields + (
            'files',
            'revision',
            'element',
            'folder_type'
        )


class DataImportFromFilesSerializer(serializers.Serializer):

    mode = EnumField(DataImportMode, default=DataImportMode.Images)
    files = serializers.PrimaryKeyRelatedField(queryset=DataFile.objects.all(), many=True)
    folder_id = serializers.UUIDField(required=False, allow_null=True)
    folder_type = serializers.SlugField(required=False, allow_null=True)
    element_type = serializers.SlugField()

    default_error_messages = {
        'mode_not_allowed': 'This mode is not allowed when importing from files',
        'files_required': 'At least one file is required to start an import',
        'unique_corpus': 'Imports can only run on files from a single corpus',
        'corpus_read_only': 'Cannot write in corpus',
        'folder_not_found': 'Folder does not exist',
        'unique_pdf': 'Cannot import more than one PDF at once',
        'pdf_only': 'PDF imports can only import PDF files',
        'images_only': 'Image imports can only import image files',
        'iiif_only': 'IIIF imports can only import IIIF documents',
        'folder_required': 'Either folder_type, folder_id or both are required',
        'iiif_folder_required': 'IIIF imports require both folder_type and element_type',
        'type_not_found': 'Element type {slug!r} does not exist',
        'type_not_folder': 'Element type {slug!r} is not a folder',
        'type_folder': 'Element type {slug!r} is a folder',
    }

    def validate_mode(self, mode):
        if mode not in (DataImportMode.Images, DataImportMode.PDF, DataImportMode.IIIF):
            self.fail('mode_not_allowed')
        return mode

    def validate_files(self, files):
        if not files:
            self.fail('files_required')
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            self.fail('unique_corpus')
        corpus = corpora.pop()
        if corpus not in Corpus.objects.writable(self.context['request'].user):
            self.fail('corpus_read_only')
        return files

    def validate_folder_id(self, folder_id):
        if not folder_id:
            return None
        try:
            return Element.objects.get(
                id=folder_id,
                type__folder=True,
            )
        except Element.DoesNotExist:
            self.fail('folder_not_found')

    def validate(self, data):
        if data['mode'] == DataImportMode.PDF:
            if len(data['files']) > 1:
                self.fail('unique_pdf')
            if data['files'][0].content_type != 'application/pdf':
                self.fail('pdf_only')

        elif data['mode'] == DataImportMode.Images:
            if not all(f.content_type.startswith('image/') for f in data['files']):
                self.fail('images_only')

        elif data['mode'] == DataImportMode.IIIF:
            if not set(f.content_type for f in data['files']) <= {'application/json', 'application/ld+json'}:
                self.fail('iiif_only')

        else:
            raise NotImplementedError

        element_type_slug = data.get('element_type')
        element_type = data['files'][0].corpus.types.filter(slug=element_type_slug).first()
        if not element_type:
            self.fail('type_not_found', slug=element_type_slug)
        if element_type.folder:
            self.fail('type_folder', slug=element_type_slug)

        data['element_type'] = element_type
        folder_type_slug, folder_id = data.get('folder_type'), data.get('folder_id')

        if folder_type_slug:
            folder_type = data['files'][0].corpus.types.filter(slug=folder_type_slug).first()
            if not folder_type:
                self.fail('type_not_found', slug=folder_type_slug)
            if not folder_type.folder:
                self.fail('type_not_folder', slug=folder_type_slug)
            data['folder_type'] = folder_type
        else:
            if data['mode'] == DataImportMode.IIIF:
                # folder_type is required in IIIF
                self.fail('iiif_folder_required')
            if not folder_id:
                # Either folder_type or folder_id are required for other imports
                self.fail('folder_required')

        return data


class StartProcessSerializer(serializers.Serializer):
    chunks = serializers.IntegerField(min_value=1, max_value=10, default=1)
    thumbnails = serializers.BooleanField(default=False)


class ElementsWorkflowSerializer(serializers.Serializer):

    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())

    # Elements filtering options
    element = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none(), required=False)
    name = serializers.CharField(required=False)
    type = serializers.SlugField(required=False)
    best_class = BestClassField(required=False)
    selection = serializers.BooleanField(default=False)
    load_children = serializers.BooleanField(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora
        # Only elements representing a folder or containing a zone can be imported
        self.fields['element'].queryset = Element.objects.filter(
            Q(type__folder=True) | Q(zone__isnull=False),
            corpus__in=corpora
        )

    def validate(self, data):
        """
        Handle filters validation
        """
        corpus = data.get('corpus')
        targeted_elts = corpus.elements

        element = data.get('element')
        name = data.get('name')
        elt_type = data.get('type')
        best_class = data.get('best_class')
        selection = data.get('selection')

        if selection and not settings.ARKINDEX_FEATURES['selection']:
            raise serializers.ValidationError({'selection': ['Selection is not available on this instance.']})

        if selection and (element or name or elt_type or best_class is not None):
            raise serializers.ValidationError({'__all__': [
                'Filtering parameters (element, name, type, best_class) cannot be used with selection parameter'
            ]})
        if element and (name or elt_type or best_class is not None):
            raise serializers.ValidationError({
                '__all__': ['Filtering parameters (name, type, best_class) cannot be used with element parameter']
            })
        if element:
            # Assert element is in the right corpus
            if corpus.id != element.corpus.id:
                raise serializers.ValidationError({
                    'element': ['Element is not part of corpus {}'.format(corpus.name)]
                })

        if elt_type:
            # Check type filter is valid
            if not corpus.types.filter(slug=elt_type).exists():
                raise serializers.ValidationError({
                    'type': ['Unrecognized element type filter']
                })
            targeted_elts = targeted_elts.filter(type__slug=elt_type)

        if best_class is not None:
            best_classification_queryset = Classification \
                .objects \
                .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True)) \
                .exclude(state=ClassificationState.Rejected)

            if isinstance(best_class, UUID):
                if not corpus.ml_classes.filter(id=best_class).exists():
                    raise serializers.ValidationError({
                        'best_class': [f'MLClass with ID {best_class} does not exist']
                    })
                targeted_elts = targeted_elts.filter(
                    classifications__in=best_classification_queryset.filter(ml_class_id=best_class),
                )
                data['best_class'] = str(data['best_class'])
            elif best_class:
                targeted_elts = targeted_elts.filter(classifications__in=best_classification_queryset)
            else:
                targeted_elts = targeted_elts.exclude(classifications__in=best_classification_queryset)

        if name:
            targeted_elts = targeted_elts.filter(name__icontains=name)

        # Assert filters match Arkindex elements
        # This check will not properly work when using the selection; this case is handled in the API endpoint.
        if not targeted_elts.exists():
            raise serializers.ValidationError({
                '__all__': ['No element match filtering']
            })
        return data


class WorkerRunLightSerializer(serializers.ModelSerializer):
    """
    Serialize a worker run
    """
    parents = serializers.ListField(child=serializers.UUIDField())
    version = serializers.PrimaryKeyRelatedField(queryset=WorkerVersion.objects.all())

    class Meta:
        model = WorkerRun
        fields = (
            'id',
            'parents',
            'version',
        )


class WorkerRunSerializer(WorkerRunLightSerializer):
    """
    Serialize a worker run
    """
    dataimport = serializers.PrimaryKeyRelatedField(queryset=DataImport.objects.all())

    class Meta(WorkerRunLightSerializer.Meta):
        fields = WorkerRunLightSerializer.Meta.fields + (
            'dataimport',
        )

    def validate(self, data):
        dataimport = data.get('dataimport')

        if dataimport:
            if dataimport.mode != DataImportMode.Workers:
                raise serializers.ValidationError({'dataimport': ['Import mode must be Workers']})

            if dataimport.workflow is not None:
                raise serializers.ValidationError({'dataimport': ["Can't update a WorkerRun with a DataImport that has already started"]})

        if self.instance.dataimport.workflow is not None:
            raise serializers.ValidationError({'__all__': ["Can't update a WorkerRun related to a DataImport that has already started"]})

        return data


class ImportTranskribusSerializer(serializers.Serializer):
    """
    Serialize a Transkribus import
    """
    collection_id = serializers.IntegerField(min_value=1)
    build_entities = serializers.BooleanField(default=False)

    def validate(self, data):
        collection_id = data.get('collection_id')

        # Check Transkribus email
        transkribus_email = self.context['request'].user.transkribus_email
        if not transkribus_email:
            raise serializers.ValidationError(
                {"__all__": ["You have not register your transkribus email"]}
            )

        # Login as Arkindex user on transkribus
        try:
            transkribus_client = TranskribusAPI(email=settings.TRANSKRIBUS_EMAIL, password=settings.TRANSKRIBUS_PASSWORD)
        except Exception:
            raise serializers.ValidationError(
                {"__all__" : [f"Failed to login on Transkribus as {settings.TRANSKRIBUS_EMAIL}"]}
            )

        # Check Arkindex's right
        try:
            users = transkribus_client.list_user_collection(collection_id)
        except Exception:
            raise serializers.ValidationError(
                {"collection_id" : [f"User {settings.TRANSKRIBUS_EMAIL} is not a member of the collection {collection_id}"]}
            )

        # Check user's right
        user = next(filter(lambda user: user["email"] == transkribus_email, users), None)
        if not user:
            raise serializers.ValidationError(
                {"collection_id": [f"User {transkribus_email} is not a member of the collection {collection_id}"]}
            )

        return data
