from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase
from arkindex.documents.models import Element, Classification, DataSource, MLClass, ClassificationState, Corpus
from arkindex.dataimport.models import DataImport, DataImportMode


class TestProcessElements(FixtureAPITestCase):
    """
    Test list process elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        # ----------------- TREE ------------------
        #
        #        Folder 1          Folder 2
        #          /|\               /|\
        #         / | \             / | \
        #        /  |  \           /  |  \
        #       /   |   \         /   |   \
        #      /    |    \       /    |    \
        #     /     |     \     /     |     \
        #    /      |      \   /      |      \
        #   /       |       \ /       |       \
        # Page 1  Page 2   Page 3  Page 4  Page 5
        #   |      |         |       |       |
        # Line 1  Line 2   Line 3  Line 4  Line 5

        # Create a corpus
        cls.private_corpus = Corpus.objects.create(name='private')

        # Create element types
        cls.folder_type, _ = cls.private_corpus.types.get_or_create(
            slug='folder',
            display_name='Folder',
            folder=True,
        )
        cls.page_type, _ = cls.private_corpus.types.get_or_create(
            slug='page',
            display_name='Page',
        )
        cls.line_type, _ = cls.private_corpus.types.get_or_create(
            slug='text_line',
            display_name='Line',
        )

        # Create 2 volumes
        cls.folder_1 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Baba au rhum",
        )
        cls.folder_2 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Mongolfiere",
        )

        # Create 2 pages in each volume and 1 common page
        cls.page_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 1",
            type=cls.page_type
        )
        cls.page_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Mongolfiere 1",
            type=cls.page_type
        )
        cls.page_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Common",
            type=cls.page_type
        )
        cls.page_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Mongolfiere 2",
            type=cls.page_type
        )
        cls.page_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 2",
            type=cls.page_type
        )
        cls.page_1.add_parent(cls.folder_1)
        cls.page_2.add_parent(cls.folder_1)
        cls.page_3.add_parent(cls.folder_1)
        cls.page_4.add_parent(cls.folder_2)
        cls.page_5.add_parent(cls.folder_2)
        cls.page_3.add_parent(cls.folder_2)

        # Create 1 lines in each page
        cls.line_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 1",
            type=cls.line_type
        )
        cls.line_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 2",
            type=cls.line_type
        )
        cls.line_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 3",
            type=cls.line_type
        )
        cls.line_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 4",
            type=cls.line_type
        )
        cls.line_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 5",
            type=cls.line_type
        )
        cls.line_1.add_parent(cls.page_1)
        cls.line_2.add_parent(cls.page_2)
        cls.line_3.add_parent(cls.page_3)
        cls.line_4.add_parent(cls.page_4)
        cls.line_5.add_parent(cls.page_5)

        # Create best classes
        source = DataSource.objects.first()
        cls.coffee_source = MLClass.objects.create(name='C0FFEE', corpus=cls.private_corpus)
        cls.food_source = MLClass.objects.create(name='F00D', corpus=cls.private_corpus)
        Classification.objects.create(
            element=cls.folder_2,
            state=ClassificationState.Validated,
            ml_class=cls.food_source,
            source=source
        )
        Classification.objects.create(
            element=cls.page_1,
            state=ClassificationState.Validated,
            ml_class=cls.coffee_source,
            source=source
        )
        Classification.objects.create(
            element=cls.page_2,
            high_confidence=True,
            ml_class=cls.food_source,
            source=source
        )
        Classification.objects.create(
            element=cls.page_3,
            state=ClassificationState.Validated,
            ml_class=cls.food_source,
            source=source
        )
        Classification.objects.create(
            element=cls.page_5,
            high_confidence=True,
            ml_class=cls.food_source,
            source=source
        )
        Classification.objects.create(
            element=cls.page_5,
            state=ClassificationState.Validated,
            ml_class=cls.coffee_source ,
            source=source
        )

    def test_requires_login(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Images,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id
        )
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_id(self):
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_no_access(self):
        corpus = Corpus.objects.create(name='private')
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Images,
            corpus_id=corpus.id,
            creator_id=self.superuser.id
        )
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_filter_elements_wrong_corpus(self):
        corpus = Corpus.objects.create(name='private')
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=corpus.id,
            creator_id=self.superuser.id,
        )
        dataimport.elements.add(self.page_1.id, self.folder_2.id)
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {'elements': [f'Some elements on this process are not part of corpus {corpus.id}']})

    def test_filter_elements_multiple_corpus(self):
        corpus = Corpus.objects.create(name='private')
        element = Element.objects.create(
            corpus=corpus,
            type=self.folder_type,
            name="Folder",
        )
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=corpus.id,
            creator_id=self.superuser.id,
        )
        dataimport.elements.add(element.id, self.page_1.id)
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {'elements': [f'Some elements on this process are not part of corpus {corpus.id}']})

    def test_filter_element_wrong_corpus(self):
        corpus = Corpus.objects.create(name='private')
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=corpus.id,
            creator_id=self.superuser.id,
            element=self.page_1
        )
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {'element_id': [f'Element {self.page_1.id} is not part of corpus {corpus.id}']})

    def test_filter_name(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            name_contains="rhum"
        )
        elements = [self.folder_1, self.page_1, self.page_5]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_name_from_element(self):
        dataimport = DataImport.objects.create(
            creator_id=self.superuser.id,
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            element=self.folder_1,
            name_contains="rhum",
            load_children=True
        )
        elements = [self.folder_1, self.page_1]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(10):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_type(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            element_type=self.folder_type
        )
        elements = [self.folder_1, self.folder_2]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_type_from_element(self):
        dataimport = DataImport.objects.create(
            creator_id=self.superuser.id,
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            element=self.folder_1,
            element_type=self.line_type,
            load_children=True
        )
        elements = [self.line_1, self.line_2, self.line_3]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_best_class_by_id(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class=self.food_source.id
        )
        elements = [self.page_5, self.page_3, self.folder_2, self.page_2]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_any_best_class(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class="true"
        )
        elements = [self.page_1, self.page_5, self.page_3, self.folder_2, self.page_2]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_no_best_class(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class="false"
        )
        elements = [self.folder_1, self.line_1, self.line_2, self.line_3, self.line_4, self.line_5, self.page_4]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_element(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            element=self.page_1
        )
        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertCountEqual(data["results"], [
            {
                'id': str(self.page_1.id),
                'type': self.page_1.type.slug,
                'name': self.page_1.name
            }
        ])

    def test_filter_elements(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
        )
        dataimport.elements.add(self.page_1.id, self.folder_2.id)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(10):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 2)
        self.assertCountEqual(data["results"], [
            {
                'id': str(self.page_1.id),
                'type': self.page_1.type.slug,
                'name': self.page_1.name
            },
            {
                'id': str(self.folder_2.id),
                'type': self.folder_2.type.slug,
                'name': self.folder_2.name
            }
        ])

    def test_any_filter(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
        )
        elements = Element.objects.filter(corpus=self.private_corpus).order_by('name', 'type__slug')
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], elements.count())
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_name(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            name_contains="rhum",
            load_children=True
        )
        elements = [self.folder_1, self.page_1, self.page_5]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_corpus_and_filter_type(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            element_type=self.folder_type,
            load_children=True
        )
        elements = [self.folder_1, self.folder_2]
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_best_class_by_id(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class=self.food_source.id,
            load_children=True

        )
        elements = [self.folder_2, self.page_2, self.page_3, self.page_5]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_best_class(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class="true",
            load_children=True

        )
        elements = [self.page_1, self.page_5, self.page_3, self.folder_2, self.page_2]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_no_best_class(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            best_class="false",
            load_children=True
        )
        elements = [self.folder_1, self.line_1, self.line_2, self.line_3, self.line_4, self.line_5, self.page_4]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_element(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            element=self.folder_1,
            load_children=True
        )
        elements = [self.folder_1, self.page_1, self.page_3, self.line_1, self.line_2, self.line_3, self.page_2]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(10):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_elements(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            load_children=True
        )
        dataimport.elements.add(self.page_1.id, self.folder_2.id)
        elements = [self.page_1, self.page_5, self.page_3, self.line_1, self.line_3, self.line_4, self.line_5, self.folder_2, self.page_4]
        self.client.force_login(self.superuser)
        with self.assertNumQueries(12):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], len(elements))
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_any_filter(self):
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            load_children=True
        )
        elements = Element.objects.filter(corpus=self.private_corpus).order_by('name', 'type__slug')
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], elements.count())
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_all_modes(self):
        self.client.force_login(self.superuser)
        for mode in (DataImportMode.Images, DataImportMode.PDF, DataImportMode.Repository, DataImportMode.IIIF, DataImportMode.Transkribus):
            dataimport = DataImport.objects.create(
                mode=mode,
                corpus_id=self.private_corpus.id,
                creator_id=self.superuser.id,
                load_children=True
            )
            with self.assertNumQueries(3):
                response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 0)
            self.assertCountEqual(data["results"], [])

        elements = Element.objects.filter(corpus=self.private_corpus).order_by('name', 'type__slug')

        for mode in (DataImportMode.Elements, DataImportMode.Workers):
            dataimport = DataImport.objects.create(
                mode=mode,
                corpus_id=self.private_corpus.id,
                creator_id=self.superuser.id,
                load_children=True
            )
            with self.assertNumQueries(7):
                response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], elements.count())
            self.assertCountEqual(data["results"], [
                {
                    'id': str(element.id),
                    'type': element.type.slug,
                    'name': element.name
                }
                for element in elements
            ])

    def test_list_elements_ordering(self):
        """
        No element is duplicated or dropped
        """
        elts = Element.objects.bulk_create([
            Element(
                corpus=self.private_corpus,
                name="Similar name",
                type=self.page_type
            ) for i in range(40)
        ])
        for elt in elts:
            elt.add_parent(self.folder_1)
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Elements,
            corpus_id=self.private_corpus.id,
            creator_id=self.superuser.id,
            element=self.folder_1,
            load_children=True
        )
        self.client.force_login(self.superuser)
        with self.assertNumQueries(10):
            page_1 = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        with self.assertNumQueries(10):
            page_2 = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}), {'page': 2})
        qs_1 = Element.objects.filter(id__in=[elt['id'] for elt in page_1.json()['results']])
        qs_2 = Element.objects.filter(id__in=[elt['id'] for elt in page_2.json()['results']])

        self.assertEqual(qs_1.intersection(qs_2).count(), 0)
        self.assertEqual(qs_1.union(qs_2).distinct().count(), 40)
