from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import \
    Corpus, Entity, EntityLink, EntityRole, TranscriptionEntity
from arkindex_common.enums import EntityType
from arkindex.documents.serializers.light import CorpusLightSerializer, InterpretedDateSerializer
from arkindex.documents.serializers.ml import DataSourceSerializer
from arkindex.project.serializer_fields import EnumField, DataSourceSlugField
from arkindex.project.triggers import reindex_start


class BaseEntitySerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)
    dates = InterpretedDateSerializer(many=True, source='get_dates', read_only=True)
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    source = DataSourceSerializer(read_only=True)
    worker_version_id = serializers.PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'dates',
            'source',
            'worker_version_id',
        )
        read_only_fields = (
            'id',
            'dates',
            'source',
            'worker_version_id',
        )


class EntityRoleSerializer(serializers.ModelSerializer):
    """
    Serialize a role between two types of entity
    """
    parent_type = EnumField(EntityType)
    child_type = EnumField(EntityType)

    class Meta:
        model = EntityRole
        fields = (
            'id',
            'parent_name',
            'child_name',
            'parent_type',
            'child_type'
        )

    def validate(self, data):
        assert 'corpus' not in data
        assert self.context.get('request') is not None
        corpus_id = self.context['request'].parser_context['kwargs']['pk']
        corpus = Corpus.objects.writable(self.context['request'].user).filter(id=corpus_id).first()
        if corpus is None:
            raise serializers.ValidationError({
                'corpus': ['You do not have write access to this corpus'],
                'id': corpus_id,
            })
        data['corpus'] = corpus
        return data


class EntityLinkSerializer(serializers.ModelSerializer):
    """
    Serialize an entity link with its child, parent and role
    """
    parent = BaseEntitySerializer()
    child = BaseEntitySerializer()
    role = EntityRoleSerializer()

    class Meta:
        model = EntityLink
        fields = (
            'id',
            'parent',
            'child',
            'role'
        )


class EntitySerializer(BaseEntitySerializer):
    """
    Serialize an entity with its metadata
    """
    corpus = CorpusLightSerializer(read_only=True)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)

    class Meta:
        model = Entity
        fields = BaseEntitySerializer.Meta.fields + (
            'corpus',
            'children',
            'parents'
        )
        read_only_fields = BaseEntitySerializer.Meta.read_only_fields = (
            'corpus',
            'children',
            'parents',
        )

    def validate(self, data):
        if data.get('validated'):
            data['moderator'] = self.context['request'].user
        return data

    def update(self, instance, validated_data):
        updated_entity = super().update(instance, validated_data)
        reindex_start(entity=updated_entity, entities=True)
        return updated_entity


class EntityCreateSerializer(BaseEntitySerializer):
    """
    Serialize an entity with a possible parents and children
    """
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    metas = serializers.HStoreField(child=serializers.CharField(), required=False)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)
    ner = DataSourceSlugField(tool_type=MLToolType.NER, default=None)
    worker_version = serializers.PrimaryKeyRelatedField(queryset=WorkerVersion.objects.all(), default=None)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'corpus',
            'parents',
            'children',
            'ner',
            'worker_version'
        )
        read_only_fields = (
            'id',
            'children',
            'parents',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora

    def validate(self, data):
        ner = data.get('ner')
        worker_version = data.get('worker_version')
        if not ner and not worker_version:
            raise ValidationError({
                'ner': ['This field XOR worker_version field must be set to create an entity'],
                'worker_version': ['This field XOR ner field must be set to create an entity']
            })
        elif ner and worker_version:
            raise ValidationError({
                'ner': ['You can only refer to a DataSource XOR a WorkerVersion on an entity'],
                'worker_version': ['You can only refer to a DataSource XOR a WorkerVersion on an entity']
            })

        return data


class EntityLinkCreateSerializer(EntityLinkSerializer):
    """
    Serialize an entity with a possible parents and children
    """
    parent = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())
    child = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())
    role = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())

    class Meta:
        model = EntityLink
        fields = EntityLinkSerializer.Meta.fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        entities = Entity.objects.all().filter(corpus__in=corpora)
        roles = EntityRole.objects.all().filter(corpus__in=corpora)
        self.fields['parent'].queryset = entities
        self.fields['child'].queryset = entities
        self.fields['role'].queryset = roles

    def validate(self, data):
        data = super().validate(data)
        link = EntityLink(**data)
        link.full_clean()
        return data


class TranscriptionEntitySerializer(serializers.ModelSerializer):
    """
    Serialise the link between an entity and a transcription
    Transcription element must be passed in serializer context
    """
    offset = serializers.IntegerField(min_value=0)
    length = serializers.IntegerField(min_value=0)

    class Meta:
        model = TranscriptionEntity
        fields = ('entity', 'offset', 'length')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        self.fields['entity'].queryset = Entity.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user)
        )

    def validate(self, data):
        data = super().validate(data)
        data['transcription'] = self.context['transcription']
        if not data['transcription']:
            raise serializers.ValidationError({
                'transcription': ['invalid UUID or Corpus write-access is forbidden']
            })
        return data


class TranscriptionEntityDetailsSerializer(TranscriptionEntitySerializer):
    """
    Serialize the link between a transcription and a detailled entity
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['entity'] = BaseEntitySerializer()
