from django.urls import reverse
from django.test import override_settings
from rest_framework import status
from arkindex.documents.models import \
    Corpus, Element, Classification, ClassificationState, DataSource, MLClass
from arkindex_common.ml_tool import MLToolType
from arkindex.project.tests import FixtureAPITestCase


class TestClasses(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name='text', corpus=cls.corpus)
        cls.cover = MLClass.objects.create(name='cover', corpus=cls.corpus)
        cls.classified = cls.corpus.types.create(slug='classified', folder=True)

    def populate_classified_elements(self):
        self.folder_type = self.corpus.types.create(slug='folder', folder=True)
        self.parent = self.corpus.elements.create(type=self.folder_type)
        self.common_children = self.corpus.elements.create(type=self.folder_type)

        source1 = DataSource.objects.get(slug='test', type=MLToolType.Recognizer)
        source2 = DataSource.objects.create(
            type=MLToolType.Classifier,
            slug='source2',
            name='classifier',
            revision='123',
            internal=False,
        )
        for elt_num in range(1, 13):
            elt = Element.objects.create(
                name='elt_{}'.format(elt_num),
                type=self.classified,
                corpus_id=self.corpus.id
            )
            elt.add_parent(self.parent)
            self.common_children.add_parent(elt)
            for ml_class, score in zip((self.text, self.cover), (.7, .99)):
                for source in (source1, source2):
                    elt.classifications.create(
                        source_id=source.id,
                        ml_class_id=ml_class.id,
                        confidence=score,
                        high_confidence=bool(score == .99)
                    )

    def test_list(self):
        """
        Test listing results alpha-ordered
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover",
                    "nb_best": 0
                },
                {
                    "id": str(self.text.pk),
                    "name": "text",
                    "nb_best": 0
                },
            ]
        })

        # Add other classes to ensure ordering is preserved
        self.white_page = MLClass.objects.create(name='white-page', corpus=self.corpus)
        self.image = MLClass.objects.create(name='image', corpus=self.corpus)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 4,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.cover.pk),
                    "name": "cover",
                    "nb_best": 0
                },
                {
                    "id": str(self.image.pk),
                    "name": "image",
                    "nb_best": 0
                },
                {
                    "id": str(self.text.pk),
                    "name": "text",
                    "nb_best": 0
                },
                {
                    "id": str(self.white_page.pk),
                    "name": "white-page",
                    "nb_best": 0
                },
            ]
        })

    def test_unauthorized_methods(self):
        """
        Test unauthorized HTTP methods on ListCorpusMLClasses endpoint returns 405
        """
        self.client.force_login(self.user)
        for method in (self.client.put, self.client.patch, self.client.delete):
            response = method(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {})
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_list_classes_best_count(self):
        """
        Test nb_best attribute on ListCorpusMLClasses endpoint
        """
        self.populate_classified_elements()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertListEqual(
            data['results'],
            [
                {
                    "id": str(self.cover.id),
                    "name": "cover",
                    "nb_best": 24
                },
                {
                    "id": str(self.text.id),
                    "name": "text",
                    "nb_best": 0
                },
            ]
        )

    def test_list_classes_search(self):
        buffer_class = MLClass.objects.create(name='buffer overflow', corpus=self.corpus)
        self.populate_classified_elements()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}),
                data={'search': 'over'}
            )
        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertListEqual(data['results'], [
            {
                "id": str(buffer_class.id),
                "name": "buffer overflow",
                "nb_best": 0
            }, {
                "id": str(self.cover.id),
                "name": "cover",
                "nb_best": 24
            }
        ])
        response = self.client.get(
            reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}),
            data={'search': 'over buffer'}
        )
        self.assertListEqual(response.json()['results'], [{
            "id": str(buffer_class.id),
            "name": "buffer overflow",
            "nb_best": 0
        }])

    @override_settings(SEARCH_FILTER_MAX_TERMS=3)
    def test_search_filter_limit(self):
        self.corpus.ml_classes.create(name='buffer overflow')
        self.populate_classified_elements()
        self.client.force_login(self.user)

        response = self.client.get(
            reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}),
            data={'search': 'buf buf over buf flow'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response = self.client.get(
            reverse('api:corpus-classes', kwargs={'pk': self.corpus.pk}),
            data={'search': 'buf buff over buf flow'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'search': ['There cannot be more than 3 unique search terms.']})

    def test_corpus_classes_corpus_rights(self):
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name='private')
        response = self.client.post(reverse('api:corpus-classes', kwargs={'pk': private_corpus.pk}), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_classes_create_requires_login(self):
        response = self.client.post(reverse('api:corpus-classes', kwargs={'pk': self.corpus.id}), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_corpus_classes_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-classes', kwargs={'pk': self.corpus.id}), {'name': 'a'})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_class = self.corpus.ml_classes.get(name='a')
        self.assertDictEqual(response.json(), {
            'id': str(new_class.id),
            'name': 'a'
        })

    def test_corpus_classes_create_unique(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-classes', kwargs={'pk': self.corpus.id}), {'name': 'text'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['The fields name, corpus must make a unique set.']
        })

    def test_list_elements_db_queries(self):
        self.populate_classified_elements()
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:elements'), data={'type': self.classified.slug})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 12)

    def test_list_elements_best_classes(self):
        self.populate_classified_elements()
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements'),
                data={'type': self.classified.slug, 'with_best_classes': True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        for elt in data['results']:
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )

    def test_list_elements_best_classes_false(self):
        self.populate_classified_elements()
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements'),
                data={
                    'type': self.classified.slug,
                    'with_best_classes': False,
                }
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for elt in response.json()['results']:
            self.assertIsNone(elt['best_classes'])

    def test_element_parents_best_classes(self):
        self.populate_classified_elements()
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
                data={'type': self.classified.slug, 'with_best_classes': 1}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        for elt in data['results']:
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )

    def test_element_children_best_classes(self):
        self.populate_classified_elements()
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.parent.id)}),
                data={'type': self.classified.slug, 'with_best_classes': 'yes'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        for elt in data['results']:
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )

    def test_rejected_best_classes(self):
        """
        A machine classification that have been rejected by a human must not appear
        """
        self.populate_classified_elements()
        child = Element.objects.filter(type=self.classified.id).first()
        child.classifications.all().update(state=ClassificationState.Rejected)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.parent.id)}),
                data={'type': self.classified.slug, 'with_best_classes': 'yes'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        for elt in data['results']:
            if elt['id'] == str(child.id):
                self.assertListEqual(elt['best_classes'], [])
                continue
            self.assertNotEqual(elt['best_classes'], [])

    def test_validated_non_best_class(self):
        """
        A non best class validated by a human is considered as best as it is for the human
        """
        self.populate_classified_elements()
        parent = Element.objects.get_ascending(self.common_children.id).last()
        parent.classifications.all().filter(confidence=.7).update(state=ClassificationState.Validated)
        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
            data={'type': self.classified.slug, 'with_best_classes': 'yes'}
        )
        for elt in response.json()['results']:
            if elt['id'] == str(parent.id):
                self.assertCountEqual(
                    list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                    [('source2', .99), ('test', .99), ('source2', .7), ('test', .7)]
                )
                continue
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )

    def test_rejected_human_class(self):
        """
        A manual classification rejected by a human may not appear in best classes
        """
        self.populate_classified_elements()
        data_source, _ = DataSource.objects.get_or_create(type=MLToolType.NER, slug="manual", internal=False)
        element = Element.objects.filter(type=self.classified.id).first()
        classif = element.classifications.create(
            source_id=data_source.id,
            ml_class_id=self.text.id,
            confidence=1,
            high_confidence=True,
        )
        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
            data={'type': self.classified.slug, 'with_best_classes': 1}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for elt in response.json()['results']:
            if elt['id'] == str(element.id):
                self.assertCountEqual(
                    list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                    [('manual', 1.0), ('test', .99), ('source2', .99)]
                )
                continue
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )
        # Reject the manual classification
        classif.state = ClassificationState.Rejected
        classif.save()
        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
            data={'type': self.classified.slug, 'with_best_classes': 1}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for elt in response.json()['results']:
            self.assertCountEqual(
                list(map(lambda c: (c['source']['slug'], c['confidence']), elt['best_classes'])),
                [('source2', .99), ('test', .99)]
            )

    def test_class_filter_list_elements(self):
        self.populate_classified_elements()
        element = Element.objects.filter(type=self.classified.id).first()
        element.classifications.create(
            source_id=DataSource.objects.create(type=MLToolType.NER, slug='ner', internal=False).id,
            ml_class_id=self.text.id,
            confidence=.1337,
            high_confidence=True,
        )
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements'),
                data={'type': self.classified.slug, 'best_class': str(self.text.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))

    def test_class_filter_list_parents(self):
        self.populate_classified_elements()
        parent = Element.objects.get_ascending(self.common_children.id).last()
        parent.classifications.all().filter(confidence=.7).update(state=ClassificationState.Validated)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
                data={'type': self.classified.slug, 'best_class': str(self.text.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(parent.id))

    def test_class_filter_list_children(self):
        self.populate_classified_elements()
        child = Element.objects.filter(type=self.classified.id).first()
        child.classifications.all().filter(confidence=.7).update(state=ClassificationState.Validated)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.parent.id)}),
                data={'type': self.classified.slug, 'best_class': str(self.text.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(child.id))

    def test_class_filter_list_elements_distinct(self):
        self.populate_classified_elements()
        self.assertEqual(Classification.objects.filter(high_confidence=True).count(), 24)
        self.assertEqual(Classification.objects.filter(high_confidence=True).distinct('element_id').count(), 12)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements'),
                data={'type': self.classified.slug, 'best_class': str(self.cover.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        # Ensure element IDs are unique
        ids = [e['id'] for e in data['results']]
        self.assertCountEqual(ids, set(ids))

    def test_class_filter_list_parents_distinct(self):
        self.populate_classified_elements()
        self.assertEqual(Classification.objects.filter(high_confidence=True).count(), 24)
        self.assertEqual(Classification.objects.filter(high_confidence=True).distinct('element_id').count(), 12)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(self.common_children.id)}),
                data={'type': self.classified.slug, 'best_class': str(self.cover.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        # Ensure element IDs are unique
        ids = [e['id'] for e in data['results']]
        self.assertCountEqual(ids, set(ids))

    def test_class_filter_list_children_distinct(self):
        self.populate_classified_elements()
        self.assertEqual(Classification.objects.filter(high_confidence=True).count(), 24)
        self.assertEqual(Classification.objects.filter(high_confidence=True).distinct('element_id').count(), 12)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.parent.id)}),
                data={'type': self.classified.slug, 'best_class': str(self.cover.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        # Ensure element IDs are unique
        ids = [e['id'] for e in data['results']]
        self.assertCountEqual(ids, set(ids))

    def test_class_filter_true(self):
        self.populate_classified_elements()
        element = Element.objects.filter(type=self.classified.id).first()
        element.classifications.all().delete()
        element.classifications.create(
            source_id=DataSource.objects.create(type=MLToolType.NER, slug='ner', internal=False).id,
            ml_class_id=self.text.id,
            confidence=.1337,
            high_confidence=True,
        )
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements'),
                data={'type': self.classified.slug, 'best_class': 'true', 'with_best_classes': 'true'},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 12)
        best_class_ids = set(
            best_class['ml_class']['id']
            for element in data['results']
            for best_class in element['best_classes']
        )
        self.assertSetEqual(best_class_ids, {str(self.text.id), str(self.cover.id)})

    def test_class_filter_false(self):
        self.populate_classified_elements()
        element = Element.objects.filter(type=self.classified.id).first()
        element.classifications.all().delete()
        element.classifications.create(
            source_id=DataSource.objects.create(type=MLToolType.NER, slug='ner', internal=False).id,
            ml_class_id=self.text.id,
            confidence=.1337,
            high_confidence=False,
        )
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements'),
                data={'type': self.classified.slug, 'best_class': 'false', 'with_best_classes': 'true'},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertListEqual(data['results'][0]['best_classes'], [])
