# Generated by Django 2.2.13 on 2020-07-29 12:24

from django.db import migrations
from django.contrib.postgres.operations import CreateExtension
from arkindex.project.fields import LinearRingField
from arkindex.project.gis import SimplifyPreserveTopology
from arkindex.images.models import POLYGON_MAX_MEM_SIZE

POLYGON_HASH_FORWARD_SQL = """
DROP INDEX IF EXISTS zone_unique_image_polygon;
ALTER TABLE images_zone DROP COLUMN IF EXISTS polygon_hash;
"""

POLYGON_HASH_REVERSE_SQL = """
ALTER TABLE images_zone
    ADD COLUMN IF NOT EXISTS polygon_hash VARCHAR(32)
    GENERATED ALWAYS AS (md5(polygon::varchar)) STORED;
CREATE UNIQUE INDEX zone_unique_image_polygon ON images_zone (image_id, polygon_hash);
"""

# Experiments with production data has shown a tolerance of 15 is enough to simplify all polygons;
# we will try with tolerances from 1 to 20 to be safe.
MAX_TOLERANCE = 20


def simplify_polygons(apps, schema_editor):
    Zone = apps.get_model('images', 'Zone')
    if not Zone.objects.exists():
        return

    print('Looking for polygons exceeding the maximum allowed size…')
    count = Zone.objects.filter(polygon__memsize__gt=POLYGON_MAX_MEM_SIZE).count()
    if not count:
        return

    for tolerance in range(1, MAX_TOLERANCE - 1):
        print(f'Simplifying {count} polygons with tolerance {tolerance}…')
        Zone.objects \
            .filter(polygon__memsize__gt=POLYGON_MAX_MEM_SIZE) \
            .update(polygon=SimplifyPreserveTopology('polygon', tolerance))

        count = Zone.objects.filter(polygon__memsize__gt=POLYGON_MAX_MEM_SIZE).count()
        if not count:
            print('Simplification successful!')
            return

    raise ValueError('{count} existing polygons could not be simplified with a tolerance of {MAX_TOLERANCE}.')


class Migration(migrations.Migration):

    dependencies = [
        ('images', '0003_remove_image_datafile'),
    ]

    operations = [
        CreateExtension('postgis'),

        # In images.0001, we made Django believe we managed to get a unique constraint
        # on (image_id, polygon), so it now believes it does not need to change it.
        # However, the previous unique constraint used a hack with polygon_hash, whereas
        # the PostGIS geometry type allows an actual unique constraint without a generated column.
        # Therefore, we first drop the previous constraint, and will recreate it differently later.
        migrations.RunSQL(
            POLYGON_HASH_FORWARD_SQL,
            reverse_sql=POLYGON_HASH_REVERSE_SQL,
            state_operations=[
                migrations.RemoveConstraint(
                    model_name='zone',
                    name='zone_unique_image_polygon'
                )
            ],
        ),

        # Django is unable to migrate from a polygon to a geometry(LINESTRING)
        # since it would expect geometry(POLYGON). A PostGIS polygon has multiple rings,
        # but we can only handle one ring, so a closed LineString is what we want.
        # To update properly, we apply ST_Normalize and ST_RemoveRepeatedPoints
        # (Polygon.reorder but for PostGIS), then we add potentially missing final points
        # (ABCD instead of ABCDA), which would make a LineString be left open.
        migrations.RunSQL(
            sql="""
            ALTER TABLE images_zone
            ALTER COLUMN polygon
            TYPE geometry(LINESTRING, 0)
            USING ST_RemoveRepeatedPoints(ST_ExteriorRing(ST_Normalize(polygon::geometry(POLYGON, 0))));

            UPDATE images_zone
            SET polygon = ST_AddPoint(polygon, ST_StartPoint(polygon))
            WHERE NOT ST_IsClosed(polygon);
            """,
            reverse_sql="""
            ALTER TABLE images_zone
            ALTER COLUMN polygon
            TYPE polygon
            USING pclose(polygon::path)::polygon
            """,
            state_operations=[
                migrations.AlterField(
                    model_name='zone',
                    name='polygon',
                    field=LinearRingField(srid=0),
                )
            ]
        ),

        migrations.RunPython(
            simplify_polygons,
            reverse_code=migrations.RunPython.noop,
        ),

        # ST_Normalize, the extra ending point and simplifications can lead to a small portion
        # of the zones being duplicated due to that applying a more effective reordering than we did;
        # this can require running `arkindex deduplicate_zones` to manually deduplicate zones
        # before adding constraints.
        # Adding the constraints is also done in another migration due to 'pending trigger events'
        # on images_zone after the deduplication and simplification.
    ]
