from unittest import TestCase
from unittest.mock import patch
from enum import Enum
from io import StringIO
from pathlib import Path
from arkindex_common.config_parser import ConfigurationError, ConfigParser
from arkindex.project.config import get_settings_parser
import yaml

SAMPLES = Path(__file__).resolve().parent / 'config_samples'


class TestConfig(TestCase):

    def _dump_settings(self, data):
        """
        Dump settings as a YAML string, but turn non-primitive YAML types into their string representation.
        """
        stream = StringIO()
        dumper = yaml.SafeDumper(stream)

        def str_representer(self, data):
            if isinstance(data, Enum):
                data = data.value
            if data is None:
                return self.represent_none(data)
            elif isinstance(data, (bool, int, float, bytes, str)):
                return self.represent_data(data)
            return self.represent_str(str(data))

        dumper.add_representer(None, str_representer)
        dumper.ignore_aliases = lambda *args: True

        try:
            dumper.open()
            dumper.represent(data)
            dumper.close()
        finally:
            dumper.dispose()
        return stream.getvalue()

    # Ignore non-existent paths
    @patch('arkindex_common.config_parser.dir_path', new=Path)
    @patch('arkindex_common.config_parser.file_path', new=Path)
    def test_settings_defaults(self):
        parser = get_settings_parser(Path('/somewhere/backend/arkindex'))
        self.assertIsInstance(parser, ConfigParser)
        data = parser.parse_data({})

        with (SAMPLES / 'defaults.yaml').open() as f:
            expected = f.read()

        actual = self._dump_settings(data)

        self.maxDiff = None
        self.assertEqual(expected, actual)

    @patch('arkindex_common.config_parser.dir_path', new=Path)
    @patch('arkindex_common.config_parser.file_path', new=Path)
    def test_settings_override(self):
        parser = get_settings_parser(Path('/somewhere/backend/arkindex'))
        self.assertIsInstance(parser, ConfigParser)
        data = parser.parse(SAMPLES / 'override.yaml')

        with (SAMPLES / 'override.yaml').open() as f:
            expected = f.read()

        actual = self._dump_settings(data)

        self.maxDiff = None
        self.assertEqual(expected, actual)

    @patch('arkindex_common.config_parser._all_checks')
    def test_settings_errors(self, all_checks_mock):
        all_checks_mock.return_value = True
        parser = get_settings_parser(Path('/somewhere/backend/arkindex'))
        self.assertIsInstance(parser, ConfigParser)
        with self.assertRaises(ConfigurationError) as e:
            parser.parse(SAMPLES / 'errors.yaml')

        with (SAMPLES / 'expected_errors.yaml').open() as f:
            expected_errors = yaml.safe_load(f.read())
        self.maxDiff = None
        self.assertDictEqual(expected_errors, e.exception.errors)
